/*
We suppose that we have a device connected on the COM4 serial port and that sends "Hello" if we send the 
integer 63 (ASCII code of '?'), with a 8N1 configuration at 9600 bps. 
Note that this device could be a computer with the right program on it (similar to this one).
*/

// Standard library header to be able to use printf() and memset().
#include <stdio.h>

// Main Win32 API header.
#include <windows.h>

// Size of the buffers used to read and write data to the serial port.
#define BUFFER_SIZE 256

int main()
{
	HANDLE hDevID = INVALID_HANDLE_VALUE; // Identifier of the serial port.
	DCB dcb; // Structure defining port options.
	unsigned char writebuf[BUFFER_SIZE]; // Buffer used to write data.
	int writebuflen = 0; // Number of bytes to write.
	int writtenbytes = 0; // Number of bytes successfully written by WriteFile().
	unsigned char readbuf[BUFFER_SIZE]; // Buffer used to read data.
	int readbuflen = 0; // Number of bytes to read.
	int readbytes = 0; // Final number of bytes successfully read by all the calls of ReadFile().
	int nbbytes = 0; // Number of bytes successfully read by each call of ReadFile().
	int i = 0;

	//------------------------------------------------------------------------------------------------------
	// Step 1 : Opening the serial port
	//------------------------------------------------------------------------------------------------------

	// Open the COM4 serial port.
	hDevID = CreateFile( 
		"\\\\.\\COM4",
		GENERIC_READ | GENERIC_WRITE,
		0,    // Must be opened with exclusive-access.
		NULL, // No security attributes.
		OPEN_EXISTING, // Must use OPEN_EXISTING.
		0,    // Not overlapped I/O.
		NULL  // hTemplate must be NULL for comm devices.
		);

	if (hDevID == INVALID_HANDLE_VALUE)
	{
		printf("CreateFile() failed\n");
		return -1;
	}

	//------------------------------------------------------------------------------------------------------
	// Step 2 : Setting the serial port options according to the device specifications
	//------------------------------------------------------------------------------------------------------

	memset(&dcb, 0, sizeof(DCB));

	// Get the port options.
	if (!GetCommState(hDevID, &dcb))
	{
		printf("GetCommState() failed\n");
		return -1;
	}

	// Modify some options. See the MSDN documentation for more information.
	dcb.BaudRate = 9600;
	dcb.Parity = NOPARITY;
	dcb.ByteSize = 8;
	dcb.StopBits = ONESTOPBIT;

	// Apply the new options.
	if (!SetCommState(hDevID, &dcb))
	{
		printf("SetCommState() failed\n");
		return -1;
	}

	//------------------------------------------------------------------------------------------------------
	// Step 3 : Sending a request to the device
	//------------------------------------------------------------------------------------------------------

	// Prepare the data to send. We want the device to answer "Hello", so we need to send a '?' 
	// (character of ASCII code 63).
	writebuf[0] = 63; 
	// Sometimes the hexadecimal notation is used in this context :  
	//writebuf[0] = 0x3F;
	// The character notation can also be used :
	//writebuf[0] = '?';
	// Tip : use the Windows calculator in scientific mode to be able to do conversions 
	// between decimal and hexadecimal, and use Notepad and hold the Alt key while typing 63 to see a '?'.
	writebuflen = 1; // We want to write 1 byte.
	writtenbytes = 0;

	// Send the buffer to the device.
	if (WriteFile(hDevID, writebuf, writebuflen, (LPDWORD)&writtenbytes, NULL))
	{
		if (writtenbytes == 0)
		{
			printf("WriteFile() timed out\n");
			return -1;
		}
	}
	else
	{
		printf("WriteFile() failed\n");
		return -1;
	}

	//------------------------------------------------------------------------------------------------------
	// Step 4 : Getting the answer of the device
	//------------------------------------------------------------------------------------------------------

	memset(readbuf, 0, BUFFER_SIZE); // Fill the buffer of 0 (to erase any previous read data).
	readbuflen = 6; // We expect 6 bytes ("Hello" has 5 letters and there should be also the end-of-string character).
	readbytes = 0;
	nbbytes = 0;

	// Receive the answer of the device, that should be "Hello". Since it takes several bytes, 
	// the answer might not be received entirely with a single call of ReadFile() so 
	// we need to check the number of bytes received and retry if necessary.
	// This depends on several things, including the serial port timeout options,
	// the internal program of the device,... 
	// We might also get a blocking loop depending on these things (note that ReadFile() and WriteFile() block 
	// during the time of communication).
	while (readbytes < readbuflen)
	{
		// As readbuf might be partially filled by previous calls of ReadFile(), we must avoid overriding entirely it.
		// Therefore, we suppose that we have already read readbytes bytes and that there are data in the first 
		// readbytes bytes of readbuf. 
		if (ReadFile(hDevID, &(readbuf[readbytes]), readbuflen - readbytes, (LPDWORD)&nbbytes, NULL))
		{
			if (nbbytes == 0)
			{
				printf("ReadFile() timed out\n");
				return -1;
			}
		}
		else
		{
			printf("ReadFile() failed\n");
			return -1;
		}

		readbytes += nbbytes;
	}

	//------------------------------------------------------------------------------------------------------
	// Step 5 : Processing the answer of the device
	//------------------------------------------------------------------------------------------------------

	// Display the answer of the device as a string. This might bug if there is no end-of-string character
	// in readbuf.
	printf("%s\n", readbuf);

	// Display the answer of the device as a list of hexadecimal numbers. 
	for (i = 0; i < readbytes; i++)
	{
		printf("%.2x ", readbuf[i]);
	}
	printf("\n");

	//------------------------------------------------------------------------------------------------------
	// Step 6 : Closing the serial port
	//------------------------------------------------------------------------------------------------------

	// Close the COM4 serial port.
	if (!CloseHandle(hDevID))
	{
		printf("CloseHandle() failed\n");
		return -1;
	}

	return 0;
}
